from RSTAB.initModel import Model, clearAttributes, ConvertToDlString, deleteEmptyAttributes
from RSTAB.enums import *

class SteelBoundaryConditions():
    def __init__(self,
                no: int = 1,
                name: str = '',
                members: str = '',
                member_sets : str = '',
                intermediate_nodes : bool = False,
                different_properties_supports: bool = True,
                different_properties_hinges: bool = True,
                nodal_supports: list = [
                    [None, SteelBoundaryConditionsSupportType.SUPPORT_TYPE_FIXED_IN_Y_AND_TORSION, False, True, False, True, False, False, False,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, SteelBoundaryConditionsEccentricityTypeZ.ECCENTRICITY_TYPE_USER_VALUE, 0.0, 0.0, 0.0, ""],
                    [None, SteelBoundaryConditionsSupportType.SUPPORT_TYPE_FIXED_IN_Y_AND_TORSION, False, True, False, True, False, False, False,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, SteelBoundaryConditionsEccentricityTypeZ.ECCENTRICITY_TYPE_USER_VALUE, 0.0, 0.0, 0.0, ""]
                                        ],
                member_hinges: list = [
                    ["Start", False, False, False, False, False, False, False, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ""],
                    ["End", False, False, False, False, False, False, False, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ""]
                                      ],
                comment: str = '',
                params: dict = None,
                model = Model):

        """
        Args:
            no (int): Boundary Conditions Tag
            name (str): User Defined Boundary Conditions Name
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            intermediate_nodes (bool): Enable/Disable Intermediate Nodes Option
            different_properties_supports (bool): Different Properties Option for Supports
            different_properties_hinges (bool): Different Properties Option for Hinges
            nodal_supports (list of lists): Nodal Supports Table Definition
                nodal_supports[i][0] (int)= Node Sequence No.
                nodal_supports[i][1] (enum)= Support Type Enumeration
                nodal_supports[i][2] (bool)= Support in X Direction Option
                nodal_supports[i][3] (bool)= Support in Y Direction Option
                nodal_supports[i][4] (bool)= Support in Z Direction Option
                nodal_supports[i][5] (bool)= Restraint About X Option
                nodal_supports[i][6] (bool)= Restraint About Y Option
                nodal_supports[i][7] (bool)= Restraint About Z Option
                nodal_supports[i][8] (bool)= Restraint Warping Option
                nodal_supports[i][9] (float)= Rotation Magnitude
                nodal_supports[i][10] (float)= Rotation About X Magnitude
                nodal_supports[i][11] (float)= Rotation About Y Magnitude
                nodal_supports[i][12] (float)= Rotation About Z Magnitude
                nodal_supports[i][13] (float)= Support Spring X
                nodal_supports[i][14] (float)= Support Spring Y
                nodal_supports[i][15] (float)= Support Spring Z
                nodal_supports[i][16] (float)= Restraint Spring About X Magnitude
                nodal_supports[i][17] (float)= Restraint Spring About Y Magnitude
                nodal_supports[i][18] (float)= Restraint Spring About Z Magnitude
                nodal_supports[i][19] (float)= Restraint Spring Warping Magnitude
                nodal_supports[i][20] (enum)= Eccentricity Type in Z Enumeration
                nodal_supports[i][21] (float)= Eccentricity in X Magnitude
                nodal_supports[i][22] (float)= Eccentricity in Y Magnitude
                nodal_supports[i][23] (float)= Eccentricity in Z Magnitude
                nodal_supports[i][24] (str)= Assigned Nodes
            member_hinges (list of lists): Member Hinges Table Definition
                member_hinges[i][0] = Node Sequence No.
                member_hinges[i][1] = Release in X Option
                member_hinges[i][2] = Release in Y Option
                member_hinges[i][3] = Release in Z Option
                member_hinges[i][4] = Release About X Option
                member_hinges[i][5] = Release About Y Option
                member_hinges[i][6] = Release About Z Option
                member_hinges[i][7] = Release Warping Option
                member_hinges[i][8] = Release Spring in X Magnitude
                member_hinges[i][9] = Release Spring in Y Magnitude
                member_hinges[i][10] = Release Spring in Z Magnitude
                member_hinges[i][11] = Release Spring About X Magnitude
                member_hinges[i][12] = Release Spring About Y Magnitude
                member_hinges[i][13] = Release Spring About Z Magnitude
                member_hinges[i][14] = Release Spring Warping Magnitude
                member_hinges[i][15] = Assigned Nodes
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Types For Steel Design Boundary Conditions
        clientObject = model.clientModel.factory.create('ns0:steel_boundary_conditions')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Boundary Conditions No.
        clientObject.no = no

        # Boundary Conditions User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Intermediate Nodes Option
        clientObject.intermediate_nodes = intermediate_nodes

        # Boundary Conditions Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Boundary Conditions Assigned Member Sets
        clientObject.member_sets = ConvertToDlString(member_sets)

        # Boundary Conditions Nodal Supports
        clientObject.nodal_supports = model.clientModel.factory.create('ns0:steel_boundary_conditions.nodal_supports')

        for i,j in enumerate(nodal_supports):
            mlvlp = model.clientModel.factory.create('ns0:steel_boundary_conditions_nodal_supports_row')
            mlvlp.no = i+1
            mlvlp.row = model.clientModel.factory.create('ns0:steel_boundary_conditions_nodal_supports')
            clearAttributes(mlvlp.row)
            mlvlp.row.node_seq_no = nodal_supports[i][0]
            mlvlp.row.support_type = nodal_supports[i][1].name
            mlvlp.row.support_in_x = nodal_supports[i][2]
            mlvlp.row.support_in_y = nodal_supports[i][3]
            mlvlp.row.support_in_z = nodal_supports[i][4]
            mlvlp.row.restraint_about_x = nodal_supports[i][5]
            mlvlp.row.restraint_about_y = nodal_supports[i][6]
            mlvlp.row.restraint_about_z = nodal_supports[i][7]
            mlvlp.row.restraint_warping = nodal_supports[i][8]
            mlvlp.row.rotation = nodal_supports[i][9]
            mlvlp.row.rotation_about_x = nodal_supports[i][10]
            mlvlp.row.rotation_about_y = nodal_supports[i][11]
            mlvlp.row.rotation_about_z = nodal_supports[i][12]
            mlvlp.row.support_spring_in_x = nodal_supports[i][13]
            mlvlp.row.support_spring_in_y = nodal_supports[i][14]
            mlvlp.row.support_spring_in_z = nodal_supports[i][15]
            mlvlp.row.restraint_spring_about_x = nodal_supports[i][16]
            mlvlp.row.restraint_spring_about_y = nodal_supports[i][17]
            mlvlp.row.restraint_spring_about_z = nodal_supports[i][18]
            mlvlp.row.restraint_spring_warping = nodal_supports[i][19]
            mlvlp.row.eccentricity_type_z_type = nodal_supports[i][20].name
            mlvlp.row.eccentricity_x = nodal_supports[i][21]
            mlvlp.row.eccentricity_y = nodal_supports[i][22]
            mlvlp.row.eccentricity_z = nodal_supports[i][23]
            mlvlp.row.nodes = nodal_supports[i][24]

            clientObject.nodal_supports.steel_boundary_conditions_nodal_supports.append(mlvlp)

        # Boundary Conditions Member Hinges
        clientObject.member_hinges = model.clientModel.factory.create('ns0:steel_boundary_conditions.member_hinges')

        for i,j in enumerate(member_hinges):
            mlvlp = model.clientModel.factory.create('ns0:steel_boundary_conditions_member_hinges_row')
            mlvlp.no = i+1
            mlvlp.row = model.clientModel.factory.create('ns0:steel_boundary_conditions_member_hinges')
            clearAttributes(mlvlp.row)
            mlvlp.row.node_seq_no = member_hinges[i][0]
            mlvlp.row.release_in_x = member_hinges[i][1]
            mlvlp.row.release_in_y = member_hinges[i][2]
            mlvlp.row.release_in_z = member_hinges[i][3]
            mlvlp.row.release_about_x = member_hinges[i][4]
            mlvlp.row.release_about_y = member_hinges[i][5]
            mlvlp.row.release_about_z = member_hinges[i][6]
            mlvlp.row.release_warping = member_hinges[i][7]
            mlvlp.row.release_spring_in_x = member_hinges[i][8]
            mlvlp.row.release_spring_in_y = member_hinges[i][9]
            mlvlp.row.release_spring_in_z = member_hinges[i][10]
            mlvlp.row.release_spring_about_x = member_hinges[i][11]
            mlvlp.row.release_spring_about_y = member_hinges[i][12]
            mlvlp.row.release_spring_about_z = member_hinges[i][13]
            mlvlp.row.release_spring_warping = member_hinges[i][14]
            mlvlp.row.nodes = member_hinges[i][15]

            clientObject.member_hinges.steel_boundary_conditions_member_hinges.append(mlvlp)

        # Boundary Conditions Different Properties Supports
        clientObject.different_properties_supports = different_properties_supports

        # Boundary Conditions Different Properties Hinges
        clientObject.different_properties_hinges = different_properties_hinges

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Steel Boundary Conditions to client model
        model.clientModel.service.set_steel_boundary_conditions(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import StabilityAnalysisSettingsAnalysisType
from RSTAB.enums import StabilityAnalysisSettingsEigenvalueMethod
from RSTAB.enums import StabilityAnalysisSettingsMatrixType

class StabilityAnalysisSettings():

    def __init__(self,
                 no: int = 1,
                 name: str = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        '''
        Args:
            no (int): Stability Analysis Setting Tag
            name (str): Stability Analysis Setting Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Stability Analysis Settings
        clientObject = model.clientModel.factory.create('ns0:stability_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Stability Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Stability Analysis Type
        clientObject.analysis_type = StabilityAnalysisSettingsAnalysisType.EIGENVALUE_METHOD.name

        # Number of Lowest Eigenvalues
        clientObject.number_of_lowest_eigenvalues = 4

        # Eigenvalue Method
        clientObject.eigenvalue_method = StabilityAnalysisSettingsEigenvalueMethod.EIGENVALUE_METHOD_SUBSPACE_ITERATION.name

        # Matrix Type
        clientObject.matrix_type = StabilityAnalysisSettingsMatrixType.MATRIX_TYPE_STANDARD.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Stability Analysis Settings to client model
        model.clientModel.service.set_stability_analysis_settings(clientObject)

    @staticmethod
    def EigenvalueMethod(
                 no: int = 1,
                 name: str = None,
                 number_of_lowest_eigenvalues: int = 4,
                 considered_favored_effect: bool = True,
                 critical_load_factor = None,
                 minimum_initial_strain = 1e-05,
                 local_torsional_rotations = None,
                 eigenvalue_method = StabilityAnalysisSettingsEigenvalueMethod.EIGENVALUE_METHOD_SUBSPACE_ITERATION,
                 matrix_type = StabilityAnalysisSettingsMatrixType.MATRIX_TYPE_STANDARD,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        '''
        Args:
            no (int): Stability Analysis Setting Tag
            name (str, optional): Stability Analysis Setting Name
            number_of_lowest_eigenvalues (int): Number of Lowest Eigenvalues
            considered_favored_effect (bool): Considered Favored Effect
            critical_load_factor (int, optional): Critical Load Factor
                for find_eigenvectors_beyond_critical_load_factor == False:
                    critical_load_factor = None
                for find_eigenvectors_beyond_critical_load_factor == True:
                    critical_load_factor = int
            minimum_initial_strain (optional): Minimum Initial Strain
                for minimum initial strain application:
                    minimum_initial_strain != 0 or minimum_initial_strain
                for no minimum initial strain application:
                    minimum_initial_strain == 0 or minimum_initial_strain is None
            local_torsional_rotations (optional): Local Torsional Rotations
                for no local torsional rotations display:
                    local_torsional_rotations = None
                for local torsional rotations display:
                    local_torsional_rotations = double
            eigenvalue_method (enum): StabilityAnalysisSettings Eigenvalue Method Enumeration
            matrix_type (enum): StabilityAnalysisSettings Matrix Type Enumeration
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Stability Analysis Settings
        clientObject = model.clientModel.factory.create('ns0:stability_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Stability Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Stability Analysis Type
        clientObject.analysis_type = StabilityAnalysisSettingsAnalysisType.EIGENVALUE_METHOD.name

        # Number of Lowest Eigenvalues
        clientObject.number_of_lowest_eigenvalues = number_of_lowest_eigenvalues

        # Considered Favored Effect
        clientObject.considered_favored_effect = considered_favored_effect

        # Finding Eigenvectors Beyond Critical Load Factor
        if critical_load_factor:
            clientObject.find_eigenvectors_beyond_critical_load_factor = True
            clientObject.critical_load_factor = critical_load_factor

        # Minimum Initial Strain
        if minimum_initial_strain in (None, 0):
            clientObject.activate_minimum_initial_prestress = False
        else:
            clientObject.activate_minimum_initial_prestress = True
            clientObject.minimum_initial_strain = minimum_initial_strain

        # Local Torsional Relations
        if local_torsional_rotations:
            clientObject.display_local_torsional_rotations = True

        # Eigenvalue Method
        clientObject.eigenvalue_method = eigenvalue_method.name

        # Matrix Type
        clientObject.matrix_type = matrix_type.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Stability Analysis Settings to client model
        model.clientModel.service.set_stability_analysis_settings(clientObject)

    @staticmethod
    def IncrementalyMethodWithEigenvalue(
                 no: int = 1,
                 name: str = None,
                 number_of_lowest_eigenvalues: int = 4,
                 considered_favored_effect: bool = True,
                 critical_load_factor = None,
                 minimum_initial_strain = 1e-05,
                 local_torsional_rotations = None,
                 incrementally_increasing_loading = [1.0, 0.1, 10, 100],
                 stopping_of_load_increasing = None,
                 save_results_of_all_increments: bool = False,
                 eigenvalue_method = StabilityAnalysisSettingsEigenvalueMethod.EIGENVALUE_METHOD_SUBSPACE_ITERATION,
                 matrix_type = StabilityAnalysisSettingsMatrixType.MATRIX_TYPE_STANDARD,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        '''
        Args:
            no (int): Stability Analysis Setting Tag
            name (str, optional): Stability Analysis Setting Name
            number_of_lowest_eigenvalues (int): Number of Lowest Eigenvalues
            considered_favored_effect (bool): Considered Favored Effect
            critical_load_factor (int, optional): Critical Load Factor
                for find_eigenvectors_beyond_critical_load_factor == False:
                    critical_load_factor = None
                for find_eigenvectors_beyond_critical_load_factor == True:
                    critical_load_factor = int
            minimum_initial_strain (optional): Minimum Initial Strain
                for minimum initial strain application:
                    minimum_initial_strain != 0 or minimum_initial_strain
                for no minimum initial strain application:
                    minimum_initial_strain is None
            local_torsional_rotations (optional): Local Torsional Rotations
                for no local torsional rotations display:
                    local_torsional_rotations = None
                for local torsional rotations display:
                    local_torsional_rotations = double
            incrementally_increasing_loading (list): Incrementally Increasing Loading
                incrementally_increasing_loading = [initial_load_factor, load_factor_increment, refinement_of_the_last_load_increment, maximum_number_of_load_increments]
            stopping_of_load_increasing (list, optional): Stopping of Load Increasing
                for stopping of load increasing deactivated:
                    stopping_of_load_increasing = None
                for result u:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U, limit_result_displacement, limit_node]
                for result uX:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U_X, limit_result_displacement, limit_node]
                for result uY:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U_Y, limit_result_displacement, limit_node]
                for result uZ:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U_Z, limit_result_displacement, limit_node]
                for result phi:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI, limit_result_rotation, limit_node]
                for result phiX:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI_X, limit_result_rotation, limit_node]
                for result phiY:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI_Y, limit_result_rotation, limit_node]
                for result phiZ:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI_Z, limit_result_rotation, limit_node]
            save_results_of_all_increments (bool, optional): Save Results of All Increments
            eigenvalue_method (enum): StabilityAnalysisSettings Eigenvalue Method Enumeration
            matrix_type (enum): StabilityAnalysisSettings Matrix Type Enumeration
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Stability Analysis Settings
        clientObject = model.clientModel.factory.create('ns0:stability_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Stability Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Stability Analysis Type
        clientObject.analysis_type = StabilityAnalysisSettingsAnalysisType.INCREMENTALY_METHOD_WITH_EIGENVALUE.name

        # Number of Lowest Eigenvalues
        clientObject.number_of_lowest_eigenvalues = number_of_lowest_eigenvalues

        # Considered Favored Effect
        clientObject.considered_favored_effect = considered_favored_effect

        # Finding Eigenvectors Beyond Critical Load Factor
        if critical_load_factor:
            clientObject.find_eigenvectors_beyond_critical_load_factor = True
            clientObject.critical_load_factor = critical_load_factor

        # Minimum Initial Strain
        if minimum_initial_strain in (None, 0):
            clientObject.activate_minimum_initial_prestress = False
        else:
            clientObject.activate_minimum_initial_prestress = True
            clientObject.minimum_initial_strain = minimum_initial_strain

        # Local Torsional Relations
        if local_torsional_rotations:
            clientObject.display_local_torsional_rotations = True

        # Increase Loading
        if len(incrementally_increasing_loading) != 4:
            raise ValueError('WARNING: The incrementally increasing loading parameter needs to be of length 4. Kindly check list inputs for completeness and correctness.')
        clientObject.initial_load_factor = incrementally_increasing_loading[0]
        clientObject.load_factor_increment = incrementally_increasing_loading[1]
        clientObject.refinement_of_the_last_load_increment = incrementally_increasing_loading[2]
        clientObject.maximum_number_of_load_increments = incrementally_increasing_loading[3]

        # Stopping of Load-Increasing
        if stopping_of_load_increasing:
            if len(stopping_of_load_increasing) != 3:
                raise ValueError('WARNING: For active stopping of load-increasing, the stopping of load increasing parameter needs to be of length 3. Kindly check list inputs for completeness and correctness.')
            clientObject.activate_stopping_of_load_increasing = True
            clientObject.stopping_of_load_increasing_result = stopping_of_load_increasing[0].name
            if stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U_X' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U_Y' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U_Z':
                clientObject.stopping_of_load_increasing_limit_result_displacement = stopping_of_load_increasing[1]
            elif stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI_X' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI_Y' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI_Z':
                clientObject.stopping_of_load_increasing_limit_result_rotation = stopping_of_load_increasing[1]
            clientObject.stopping_of_load_increasing_limit_node = stopping_of_load_increasing[2]

        # Save Results of All Increments
        clientObject.save_results_of_all_increments = save_results_of_all_increments

        # Eigenvalue Method
        clientObject.eigenvalue_method = eigenvalue_method.name

        # Matrix Type
        clientObject.matrix_type = matrix_type.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Stability Analysis Settings to client model
        model.clientModel.service.set_stability_analysis_settings(clientObject)

    @staticmethod
    def IncrementalyMethodWithoutEigenvalue(
                 no: int = 1,
                 name: str = None,
                 minimum_initial_strain = 1e-05,
                 local_torsional_rotations = None,
                 incrementally_increasing_loading = [1.0, 0.1, 10, 100],
                 stopping_of_load_increasing = None,
                 save_results_of_all_increments: bool = False,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        '''
        Args:
            no (int): Stability Analysis Setting Tag
            name (str, optional): Stability Analysis Setting Name
            minimum_initial_strain (optional): Minimum Initial Strain
                for minimum initial strain application:
                    minimum_initial_strain != 0 or minimum_initial_strain
                for no minimum initial strain application:
                    minimum_initial_strain == 0 or minimum_initial_strain is None
            local_torsional_rotations (optional): Local Torsional Rotations
                for no local torsional rotations display:
                    local_torsional_rotations = None
                for local torsional rotations display:
                    local_torsional_rotations = double
            incrementally_increasing_loading (list): Incrementally Increasing Loading
                incrementally_increasing_loading = [initial_load_factor, load_factor_increment, refinement_of_the_last_load_increment, maximum_number_of_load_increments]
            stopping_of_load_increasing (list, optional): Stopping of Load Increasing
                for stopping of load increasing deactivated:
                    stopping_of_load_increasing = None
                for result u:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U, limit_result_displacement, limit_node]
                for result uX:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U_X, limit_result_displacement, limit_node]
                for result uY:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U_Y, limit_result_displacement, limit_node]
                for result uZ:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_DISPLACEMENT_U_Z, limit_result_displacement, limit_node]
                for result phi:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI, limit_result_rotation, limit_node]
                for result phiX:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI_X, limit_result_rotation, limit_node]
                for result phiY:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI_Y, limit_result_rotation, limit_node]
                for result phiZ:
                    stopping_of_load_increasing = [StabilityAnalysisSettingsStoppingOfLoadIncreasingResult.RESULT_TYPE_ROTATION_PHI_Z, limit_result_rotation, limit_node]
            save_results_of_all_increments (bool, optional): Save Results of All Increments
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Stability Analysis Settings
        clientObject = model.clientModel.factory.create('ns0:stability_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Stability Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Stability Analysis Type
        clientObject.analysis_type = StabilityAnalysisSettingsAnalysisType.INCREMENTALY_METHOD_WITHOUT_EIGENVALUE.name

        # Minimum Initial Strain
        if minimum_initial_strain in (None, 0):
            clientObject.activate_minimum_initial_prestress = False
        else:
            clientObject.activate_minimum_initial_prestress = True
            clientObject.minimum_initial_strain = minimum_initial_strain

        # Local Torsional Relations
        if local_torsional_rotations:
            clientObject.display_local_torsional_rotations = True

        # Increase Loading
        if len(incrementally_increasing_loading) != 4:
            raise ValueError('WARNING: The incrementally increasing loading parameter needs to be of length 4. Kindly check list inputs for completeness and correctness.')
        clientObject.initial_load_factor = incrementally_increasing_loading[0]
        clientObject.load_factor_increment = incrementally_increasing_loading[1]
        clientObject.refinement_of_the_last_load_increment = incrementally_increasing_loading[2]
        clientObject.maximum_number_of_load_increments = incrementally_increasing_loading[3]

        # Stopping of Load-Increasing
        if stopping_of_load_increasing:
            if len(stopping_of_load_increasing) != 3:
                raise ValueError('WARNING: For active stopping of load-increasing, the stopping of load increasing parameter needs to be of length 3. Kindly check list inputs for completeness and correctness.')
            clientObject.activate_stopping_of_load_increasing = True
            clientObject.stopping_of_load_increasing_result = stopping_of_load_increasing[0].name
            if stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U_X' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U_Y' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_DISPLACEMENT_U_Z':
                clientObject.stopping_of_load_increasing_limit_result_displacement = stopping_of_load_increasing[1]
            elif stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI_X' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI_Y' or stopping_of_load_increasing[0].name == 'RESULT_TYPE_ROTATION_PHI_Z':
                clientObject.stopping_of_load_increasing_limit_result_rotation = stopping_of_load_increasing[1]
            clientObject.stopping_of_load_increasing_limit_node = stopping_of_load_increasing[2]

        # Save Results of All Increments
        clientObject.save_results_of_all_increments = save_results_of_all_increments

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Stability Analysis Settings to client model
        model.clientModel.service.set_stability_analysis_settings(clientObject)



import os
import sys
import RSTAB.dependencies # dependency check ahead of imports
import socket
import ssl
import requests
import xmltodict
from urllib import request
from suds import WebFault
from suds.client import Client
from RSTAB.enums import ObjectTypes, ModelType, AddOn
from RSTAB.suds_requests import RequestsTransport
from suds.cache import DocumentCache
from tempfile import gettempdir
import time
from RSTAB import connectionGlobals

def connectToServer(url=connectionGlobals.url, port=connectionGlobals.port):
    """
    Function for connecting to the server - code moved to function,
    so it is not executed on import of the module
    """
    # Check server port range set in "Program Options & Settings"
    # By default range is set between 8081 ... 8089
    if connectionGlobals.connected:
        return

    print('Connecting to server...')

    # local machine url format: 'http://127.0.0.1'
    urlAndPort = f'{url}:{port}'

    # Parse the hostname from the URL
    if url.startswith('https://'):
        hostname = url[8:]  # Remove 'https://'
        context = ssl.create_default_context()
        if isinstance(connectionGlobals.verify, str):
            context.load_verify_locations(cafile=connectionGlobals.verify)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        a_socket = context.wrap_socket(sock, server_hostname=hostname)
        new_wsdl = request.urlopen(urlAndPort+'/wsdl', context=context)
    elif url.startswith('http://'):
        hostname = url[7:]  # Remove 'http://'
        a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        new_wsdl = request.urlopen(urlAndPort+'/wsdl')
    else:
        hostname = url
        a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        new_wsdl = request.urlopen(urlAndPort+'/wsdl')

    location = (hostname, int(port))

    # Check if port is listening
    result_of_check = a_socket.connect_ex(location)

    if result_of_check == 0:
        a_socket.close()
    else:
        print(f'Error: Port {urlAndPort} is not open.')
        print('Please check:')
        print('- If you have started RSTAB application at the remote destination correctly.')
        a_socket.close()
        sys.exit()

    # Delete old cache if the version or mode doesn't correlate
    connectionGlobals.cacheLoc = os.path.join(gettempdir(), 'WSDL')

    new_wsdl_data = new_wsdl.read()
    new_wsdl.close()
    new_tns = xmltodict.parse(new_wsdl_data)['definitions']['@targetNamespace']

    if os.path.exists(connectionGlobals.cacheLoc):
        for file in os.listdir(connectionGlobals.cacheLoc):
            filePath = os.path.join(connectionGlobals.cacheLoc, file)
            if file.endswith('.xml'):
                with open(filePath,'r', encoding='utf-8') as old_wsdl:
                    old_wsdl_data = old_wsdl.read()
                    old_wsdl.close()
                    old_tns = xmltodict.parse(old_wsdl_data)['definitions']['@targetNamespace']
                    if new_tns != old_tns:
                        os.remove(filePath)

    # Check for issues locally and remotely
    try:
        connectionGlobals.ca = DocumentCache(location=connectionGlobals.cacheLoc)
        trans = RequestsTransport(
            api_key=connectionGlobals.api_key,
            session=connectionGlobals.session,
            verify=connectionGlobals.verify
        )
        connectionGlobals.client = Client(urlAndPort+'/wsdl', location = urlAndPort, cache=connectionGlobals.ca, transport=trans)
        connectionGlobals.connected = True

    except Exception:
        print('Error: Connection to server failed!')
        print('Please check:')
        print('- If you have started RSTAB application')
        print('- If all RSTAB dialogs are closed')
        print('- If server port range is set correctly')
        print('- If you have a valid Web Services license')
        print('- Check Program Options & Settings > Web Services')
        print('On remote PC please check:')
        print('- If the firewall enables you to listen to selected port.')
        sys.exit()

    try:
        modelLst = connectionGlobals.client.service.get_model_list()
    except Exception:
        print('Error: Please check if all RSTAB dialogs are closed.')
        input('Press Enter to exit...')
        sys.exit()

    # Persistent connection
    # 'session' and 'trans'(port) enable Client to work within 1 session which is much faster to execute.
    # Without it the session lasts only one request which results in poor performance.
    # Assigning session to application Client (here client) instead of model Client
    # results also in poor performance.

class Model():
    clientModel = None
    clientModelDct = {}

    def __init__(self,
                 new_model: bool=True,
                 model_name: str="TestModel.rs9",
                 delete: bool=False,
                 delete_all: bool=False,
                 connect_to_server: bool=True):
        """
        Class object representing individual model in RSTAB.
        Class enables to edit multiple models in one session through holding
        one transport session active by not setting 'trans' into Client.

        Args:
            new_model (bool, optional): Set to True if new model is requested.
                If model is opened in RSTAB, FALSE should be used.
            model_name (str, optional): Defaults to "TestModel". If "" call get_active_model.
            delete (bool, optional):  Delete results
            delete_all (bool, optional): Delete all objects in Model.
        """

        # This condition is here so there is backward compatibility for test etc.
        # But it is possible now to connect to server in different place
        # and then use Model(connect_toserver=False)
        if connect_to_server:
            connectToServer()

        cModel = None
        modelLst = []
        modelVct = connectionGlobals.client.service.get_model_list()
        if modelVct:
            modelLst = modelVct.name

        # The model suffix is omitted in modelLs, so it must be omitted in model_name to match exactly
        original_model_name = model_name
        if '.' in model_name:
            model_name = model_name.split('.')[0]

        if new_model:
            # Requested new model but the model with given name is already connected
            if model_name in self.clientModelDct:
                cModel = self.clientModelDct[model_name]
                # Asuming the existing model should be recycled everything have to be deleted,
                # so the script won't add new objects on top of the old ones.
                # Mainly used in cycles.
                cModel.service.delete_all_results()
                cModel.service.delete_all()

            else:
                modelPath = ''
                # Requested new model, model with given name was NOT connected yet but file with the same name is opened in RSTAB
                if model_name in modelLst:
                    id = 0
                    for i,j in enumerate(modelLst):
                        if modelLst[i] == model_name:
                            id = i
                    modelPath =  connectionGlobals.client.service.get_model(id)

                # Requested new model, model with given name DOESN'T exist yet
                else:
                    # If name is empty, active will be selected
                    if model_name == "":
                        modelPath =  connectionGlobals.client.service.get_active_model()
                    # If there is no nodel with given name, new RSTAB model will be created
                    else:
                        modelPath =  connectionGlobals.client.service.new_model(original_model_name)

                modelPort = modelPath[-5:-1]
                modelUrlPort = connectionGlobals.url+':'+modelPort
                modelCompletePath = modelUrlPort+'/wsdl'

                connectionGlobals.session = requests.Session()
                adapter = requests.adapters.HTTPAdapter(pool_connections=1, pool_maxsize=1)
                connectionGlobals.session.mount('http://', adapter)
                connectionGlobals.session.mount('https://', adapter)
                trans = RequestsTransport(
                    api_key=connectionGlobals.api_key,
                    session = connectionGlobals.session,
                    verify=connectionGlobals.verify
                )

                cModel = Client(modelCompletePath, transport=trans, location = modelUrlPort, cache=connectionGlobals.ca, timeout=360)

                self.clientModelDct[model_name] = cModel

        else:
            # Requested model is already opened in RSTAB or even connected in self.clientModelDct.
            # In this statement RSTAB doesn't create new model in RSTAB via new_model().

            # assert model_name in self.clientModelDct or model_name in modelLst, 'WARNING: '+model_name +' is not connected neither opened in RSTAB.'

            # If model with same name is opened and alredy in clientModelDct.
            # This is typicaly model created by RSTAB Python Client.
            if model_name in self.clientModelDct:
                cModel = self.clientModelDct[model_name]
            # If opening new file.
            # Model is opened in RSTAB (model in modelLst) but it is not in clientModelDct yet to be edited or closed.
            elif model_name in modelLst:
                id = 0
                for i,j in enumerate(modelLst):
                    if modelLst[i] == model_name:
                        id = i
                modelPath =  connectionGlobals.client.service.get_model(id)
                self.clientModelDct[model_name] = cModel
                modelPort = modelPath[-5:-1]
                modelUrlPort = connectionGlobals.url+':'+modelPort
                modelCompletePath = modelUrlPort+'/wsdl'

                connectionGlobals.session = requests.Session()
                adapter = requests.adapters.HTTPAdapter(pool_connections=1, pool_maxsize=1)
                connectionGlobals.session.mount('http://', adapter)
                connectionGlobals.session.mount('https://', adapter)
                trans = RequestsTransport(
                    api_key=connectionGlobals.api_key,
                    session = connectionGlobals.session,
                    verify=connectionGlobals.verify
                )

                cModel = Client(modelCompletePath, transport=trans, location = modelUrlPort, cache=connectionGlobals.ca, timeout=360)
            elif model_name == "":
                modelPath =  connectionGlobals.client.service.get_active_model()
                modelPort = modelPath[-5:-1]
                modelUrlPort = connectionGlobals.url+':'+modelPort
                modelCompletePath = modelUrlPort+'/wsdl'

                connectionGlobals.session = requests.Session()
                adapter = requests.adapters.HTTPAdapter(pool_connections=1, pool_maxsize=1)
                connectionGlobals.session.mount('http://', adapter)
                connectionGlobals.session.mount('https://', adapter)
                trans = RequestsTransport(
                    api_key=connectionGlobals.api_key,
                    session = connectionGlobals.session,
                    verify=connectionGlobals.verify
                )

                cModel = Client(modelCompletePath, transport=trans, location = modelUrlPort, cache=connectionGlobals.ca, timeout=360)
            else:
                print('Model name "'+model_name+'" is not created in RSTAB. Consider changing new_model parameter in Model class from False to True.')
                sys.exit()

        if delete:
            print('Deleting results...')
            cModel.service.delete_all_results()
        if delete_all:
            print('Delete all...')
            cModel.service.delete_all()

        # when using multiple instances/model
        self.clientModel = cModel
        # when using only one instance/model
        Model.clientModel = cModel

    def __delete__(self, index_or_name):
        '''
        Purpose of this function is to facilitate removing client instances
        from clientModelDct dictionary, which is held in Model for the purpose of
        working with multiple models either created directly in RSTAB or opened from file.

        Args:
            index_or_name (str or int): Name of the index of model
        '''
        if isinstance(index_or_name, str) and index_or_name in self.clientModelDct:
            assert index_or_name in list(self.clientModelDct)
            self.clientModelDct.pop(index_or_name)
            if len(self.clientModelDct) > 0:
                model_key = list(self.clientModelDct)[-1]
                self.clientModel = self.clientModelDct[model_key]
            else:
                self.clientModel = None
        if isinstance(index_or_name, int):
            assert index_or_name <= len(self.clientModelDct)
            modelLs = connectionGlobals.client.service.get_model_list()
            if modelLs and (modelLs.name[index_or_name] in self.clientModelDct):
                self.clientModelDct.pop(modelLs.name[index_or_name])
                if len(self.clientModelDct) > 0:
                    model_key = list(self.clientModelDct)[-1]
                    self.clientModel = self.clientModelDct[model_key]
                else:
                    self.clientModel = None

def clearAttributes(obj):
    '''
    Clears object attributes.
    Sets all attributes to None.
    Use it whenever you create new (sub)object.

    Args:
        obj: object to clear
    '''

    # iterator
    it = iter(obj)
    for i in it:
        obj[i[0]] = None

    return obj

def deleteEmptyAttributes(obj):
    from enum import Enum
    '''
    Delete all attributes that are None for better performance.

    Args:
        obj: object to clear
    '''
    it = [] # iterator
    try:
        it = iter(obj)
    except:
        ValueError('WARNING: Object feeded to deleteEmptyAttributes function is not iterable. It is type: '+str(type(obj))+'.')

    for i in it:
        if isinstance(i, str) or isinstance(i, int) or isinstance(i, float) or isinstance(i, bool) or isinstance(i, Enum) or not isinstance(i, tuple):
            continue
        if len(i) > 2:
            i = deleteEmptyAttributes(i)
        elif i[1] is None or i[1] == "":
            delattr(obj, i[0])
        elif isinstance(i[1], str) or isinstance(i[1], int) or isinstance(i[1], float) or isinstance(i[1], bool) or isinstance(i[1], Enum):
            pass
        else:
            if isinstance(i, tuple):
                i = list(i)
                i[1] = deleteEmptyAttributes(i[1])
                i = tuple(i)
            else:
                i[1] = deleteEmptyAttributes(i[1])

    return obj

def openFile(model_path):
    '''
    Open file with a name.
    This routine primarily adds client instance into
    Model.clientModelDct which manages all connections to the models.
    New Model class instance is invoked.
    It should be used when opening a file.

    Args:
        model_path (str): Path to RSTAB9 model.
    Returns:
        model (client instance): RSTAB model instance
    '''
    assert os.path.exists(model_path)

    file_name = os.path.basename(model_path)
    connectToServer()
    connectionGlobals.client.service.open_model(model_path)
    return Model(False, file_name)

def closeModel(index_or_name, save_changes = False):
    '''
    Close any model connected to client with index or name.
    Make sure to close the first open model last.
    First model carries whole session (locking of the RSTAB).

    Args:
        index_or_name : Model Index or Name to be Close
        save_changes (bool): Enable/Disable Save Changes Option
    '''

    connectToServer()
    if isinstance(index_or_name, int):
        Model.__delete__(Model, index_or_name)
        connectionGlobals.client.service.close_model(index_or_name, save_changes)

    elif isinstance(index_or_name, str):
        if index_or_name[-4:] == '.rs9':
            index_or_name = index_or_name[:-4]

        modelLs = connectionGlobals.client.service.get_model_list().name
        if index_or_name in modelLs:
            try:
                Model.__delete__(Model, index_or_name)
                connectionGlobals.client.service.close_model(modelLs.index(index_or_name), save_changes)
            except:
                print('Model did NOT close properly.')
        else:
            print('\nINFO: Model "'+index_or_name+'" is not opened.')
    else:
        assert False, 'Parameter index_or_name must be int or string.'

def closeAllModels(save_changes = False):
    '''
    Function that closes all opened models in reverse order.

    Args:
        save_changes (bool): Enable/Disable Save Changes Option
    '''
    try:
        modelLs = connectionGlobals.client.service.get_model_list().name
        for j in reversed(modelLs):
            closeModel(j, save_changes)
    except:
        print('No models opened.')

def saveFile(model_path):
    '''
    This function saves a model in a .rs9 file.

    Args:
        index_or_name : Model Index or Name to be Close
        model_path: Path to RSTAB9 model.
    '''
    if model_path[len(model_path) - 4:len(model_path)].lower() != '.rs9':
        model_path = model_path + '.rs9'

    Model.clientModel.service.save(model_path)

def insertSpaces(lst: list):
    '''
    Add spaces between list of numbers.
    Returns list of values.
    '''
    return ' '.join(str(item) for item in lst)

def Calculate_all(skipWarnings: bool = False, model = Model):
    '''
    Calculates model.
    CAUTION: Don't use it in unit tests!
    It works when executing tests individually but when running all of them
    it causes RSTAB to stuck and generates failures, which are hard to investigate.

    Args:
        skipWarnings (bool): Warnings will be skipped
        model (RSTAB Class, optional): Model to be edited
    '''

    from RSTAB.Tools.PlausibilityCheck import PlausibilityCheck
    PlausibilityCheck()

    calculationMessages = model.clientModel.service.calculate_all(skipWarnings)
    return calculationMessages

def CalculateInCloud(machine_id, run_plausibility_check, calculate_despite_warnings_and_errors, email_notification, model = Model):
    '''
    Sends the current model to the defined server to be calculated in the cloud. Plausibility check before and email notification after the cloud calculation are optional.
    CAUTION: Don't use it in unit tests!
    It works when executing tests individually but when running all of them
    it causes RSTAB to stuck and generates failures, which are hard to investigate.

    Args:
        machine_id (str): virtual machine ID (Dlu_1, F4s_v2, F8s_v2, F16s_v2, F32s_v2)
        run_plausibility_check (bool): Activate/Deactivate plausibility check of model before cloud calculation is started
        calculate_despite_warnings_and_errors (bool): Activate/Deactivate to start cloud calculation despite warnings and errors during plausibility check
        email_notification (bool): Activate/Deactivate email notification about start and end of cloud calculation
    '''
    try:
        cloudCalculationResult = model.clientModel.service.calculate_all_in_cloud(machine_id, run_plausibility_check, calculate_despite_warnings_and_errors, email_notification)
        print("Cloud calculation was started.")
        return cloudCalculationResult # list
    except WebFault as e:
        print(f"Caught exception: {e.fault.faultstring}")

def ConvertToDlString(s):
    '''
    The function converts strings commonly used in RSTAB so that they
    can be used In WebServices. It solved issue #4.
    Examples:
    '1,3'       -> '1 3'
    '1, 3'      -> '1 3'
    '1-3'       -> '1 2 3'
    '1,3,5-9'   -> '1 3 5 6 7 8 9'

    Args:
        s (str): RSTAB Common String

    Returns a WS conform string.
    '''

    # Parameter is not of required type.
    assert isinstance(s, (list, str))

    if isinstance(s, list):
        return ' '.join(map(str, s))

    s = s.strip()
    s = s.replace(',', ' ')
    s = s.replace('  ', ' ')
    lst = s.split(' ')
    new_lst = []
    for element in lst:
        if '-' in element:
            inLst = element.split('-')
            start = int(inLst[0])
            end   = int(inLst[1])
            inLst = []
            for i in range(start, end + 1):
                inLst.append(str(i))

            inS = ' '.join(inLst)
            new_lst.append(inS)
        else:
            new_lst.append(element)

    return ' '.join(new_lst)

def ConvertStrToListOfInt(st):
    """
    This function coverts string to list of integers.
    Args:
        st (str): RSTAB Common String
    """
    st = ConvertToDlString(st)
    lstInt = []
    while st:
        intNumber = 0
        if ' ' in st:
            idx = st.index(' ')
            intNumber = int(st[:idx])
            st = st[idx+1:]
        else:
            intNumber = int(st)
            st = ''
        lstInt.append(intNumber)
    return lstInt

def CheckIfMethodOrTypeExists(modelClient, method_or_type, unitTestMode=False):
    """
    Check if SOAP method or type is present in your version of RSTAB.
    Use it only in your examples.
    Unit tests except msg from SUDS where this is checked already.

    Args:
        modelClient (Model.clientModel)
        method_or_type (str): Method or Type of SOAP Client
        unitTestMode (bool): Unit Test Mode

    Returns:
        bool: Status of method or type.

    Note:
        To get list of methods invoke:
        list_of_methods = [method for method in Model.clientModel.wsdl.services[0].ports[0]]
    """
    assert modelClient, "WARNING: modelClient is not initialized."

    if method_or_type not in str(modelClient):
        if unitTestMode:
            return True
        else:
            assert False, "WARNING: Used method/type: %s is not implemented in Web Services yet." % (method_or_type)

    return not unitTestMode

def GetAllAddonStatuses(modelClient):
    """
    Get  statuses for all Addons.

    Args:
        modelClient (Model.clientModel)

    Returns:
        (dict): Addons with their statuses as values
    """
    if modelClient is None:
        print("WARNING: modelClient is not initialized.")
        return False

    addons = modelClient.service.get_addon_statuses()
    dct = {}
    for lstType in addons:
        if not isinstance(lstType[1], bool) and len(lstType[1]) > 1:
            addon = [lst for lst in lstType[1]]
            for item in addon:
                dct[str(item[0])] = bool(item[1])
        elif isinstance(lstType[1], bool):
            dct[str(lstType[0])] = bool(lstType[1])
        else:
            assert False

    return dct

def GetAddonStatus(modelClient, addOn = AddOn.stress_analysis_active):
    """
    Check if Add-on is reachable and active.
    For some types of objects, specific Add-ons need to be enabled.

    Args:
        modelClient (Model.clientModel)
        addOn (enum): AddOn Enumeration

    Returns:
        (bool): Status of Add-on
    """
    dct = GetAllAddonStatuses(modelClient)

    # sanity check
    assert addOn.name in dct, f"WARNING: {addOn.name} Add-on can not be reached."

    return dct[addOn.name]

def SetAddonStatus(modelClient, addOn = AddOn.stress_analysis_active, status = True):
    """
    Activate or deactivate Add-on.
    For some types of objects, specific Add-ons need to be enabled.

    Args:
        modelClient (Model.clientModel)
        addOn (enum): AddOn Enumeration
        status (bool): in/active

    Analysis addOns list:
        material_nonlinear_analysis_active
        structure_stability_active
        construction_stages_active
        time_dependent_active
        influence_lines_areas_active
        form_finding_active
        cutting_patterns_active
        torsional_warping_active
        cost_estimation_active

    Design addOns list:
        stress_analysis_active
        concrete_design_active
        steel_design_active
        timber_design_active
        aluminum_design_active
        steel_joints_active
        timber_joints_active
        craneway_design_active

    Dynamic addOns list:
        modal_active
        equivalent_lateral_forces_active
        spectral_active
        time_history_active
        pushover_active
        harmonic_response_active

    Special aadOns list:
        building_model_active
        wind_simulation_active
        tower_wizard_active
        tower_equipment_wizard_active
        piping_active
        air_cushions_active
        geotechnical_analysis_active
    """

    # this will also provide sanity check
    currentStatus = GetAddonStatus(modelClient, addOn)
    if currentStatus != status:
        addonLst = modelClient.service.get_addon_statuses()
        if addOn.name in addonLst['__keylist__']:
            addonLst[addOn.name] = status
        else:
            for listType in addonLst['__keylist__']:
                if not isinstance(addonLst[listType], bool) and addOn.name in addonLst[listType]:
                    addonLst[listType][addOn.name] = status

        modelClient.service.set_addon_statuses(addonLst)

def SetAddonStatuses(AddOnDict, model = Model):
    """
    Set all or selected Add-on.

    Args:
        modelClient (Model.clientModel)
        AddOnDict (dict): AddOn Dictionary

    Returns:
        (bool): If all Add-ons were set
    """
    currentStatus = model.clientModel.service.get_addon_statuses()

    for addon in currentStatus['__keylist__']:
        if addon in AddOnDict:
            currentStatus[addon] = AddOnDict[addon]
        elif not isinstance(currentStatus[addon], bool):
            for listType in currentStatus[addon]['__keylist__']:
                if isinstance(currentStatus[addon][listType], bool) and listType in AddOnDict:
                    currentStatus[addon][listType] = AddOnDict[listType]

    model.clientModel.service.set_addon_statuses(currentStatus)


def CalculateSelectedCases(loadCases: list = None, designSituations: list = None, loadCombinations: list = None,skipWarnings = True, model = Model):
    '''
    This method calculate just selected objects - load cases, designSituations, loadCombinations

    Args:
        loadCases (list, optional): Load Case List
        designSituations (list, optional): Design Situations List
        loadCombinations (list, optional): Load Combinations List
        model (RSTAB Class, optional): Model to be edited
    '''
    specificObjectsToCalculate = model.clientModel.factory.create('ns0:calculate_specific_loadings')
    if loadCases:
        for loadCase in loadCases:
            specificObjectsToCalculateLS = model.clientModel.factory.create('ns0:calculate_specific_loadings.loading')
            specificObjectsToCalculateLS.no = loadCase
            specificObjectsToCalculateLS.type = ObjectTypes.E_OBJECT_TYPE_LOAD_CASE.name
            specificObjectsToCalculate.loading.append(specificObjectsToCalculateLS)

    if designSituations:
        for designSituation in designSituations:
            specificObjectsToCalculateDS = model.clientModel.factory.create('ns0:calculate_specific_loadings.loading')
            specificObjectsToCalculateDS.no = designSituation
            specificObjectsToCalculateDS.type = ObjectTypes.E_OBJECT_TYPE_DESIGN_SITUATION.name
            specificObjectsToCalculate.loading.append(specificObjectsToCalculateDS)


    if loadCombinations:
        for loadCombination in loadCombinations:
            specificObjectsToCalculateCC = model.clientModel.factory.create('ns0:calculate_specific_loadings.loading')
            specificObjectsToCalculateCC.no = loadCombination
            specificObjectsToCalculateCC.type = ObjectTypes.E_OBJECT_TYPE_LOAD_COMBINATION.name
            specificObjectsToCalculate.loading.append(specificObjectsToCalculateCC)

    errors_and_warnings = []
    calculationMessages = []

    try:
        calculationMessages = model.clientModel.service.calculate_specific(specificObjectsToCalculate, skipWarnings)
    except Exception as exp:
        errors_and_warnings = ["Calculation was unsuccessful: " + repr(exp)]

    if calculationMessages["errors_and_warnings"] and calculationMessages["errors_and_warnings"]["message"]:
        errors_and_warnings = ["".join([message.message_type,\
                                        ": Input field: ", message.input_field,\
                                            ", object: ", message.object,\
                                                ", current value: ", message.current_value,\
                                                    ". Message: ", message.message]) if message.message_type == "ERROR"\
                                                        else "".join([message.message_type, ": ", message.message]) if not skipWarnings else None for message in calculationMessages["errors_and_warnings"]["message"]]

    return errors_and_warnings

def FirstFreeIdNumber(memType = ObjectTypes.E_OBJECT_TYPE_MEMBER, parent_no: int = 0, model = Model):
    '''
    This method returns the next available Id Number for the selected object type

    Args:
        memType (enum): Object Type Enumeration
        parent_no (int): Object Parent Number
            Note:
            (1) A geometric object has, in general, a parent_no = 0
            (2) The parent_no parameter becomes significant for example with loads
        model (RSTAB Class, optional): Model to be edited
    '''
    return model.clientModel.service.get_first_free_number(memType.name, parent_no)

def SetModelType(model_type = ModelType.E_MODEL_TYPE_3D, model = Model):
    '''
    This method sets the model type. The model type is E_MODEL_TYPE_3D by default.

    Args:
        model_type (enum): Modal Type Enumeration. The available model types are listed below.
            ModelType.E_MODEL_TYPE_1D_X_3D
            ModelType.E_MODEL_TYPE_1D_X_AXIAL
            ModelType.E_MODEL_TYPE_2D_XY_3D
            ModelType.E_MODEL_TYPE_2D_XY_PLATE
            ModelType.E_MODEL_TYPE_2D_XZ_3D
            ModelType.E_MODEL_TYPE_2D_XZ_PLANE_STRAIN
            ModelType.E_MODEL_TYPE_2D_XZ_PLANE_STRESS
            ModelType.E_MODEL_TYPE_3D
        model (RSTAB Class, optional): Model to be edited
    '''

    model.clientModel.service.set_model_type(model_type.name)

def GetModelType(model = Model):
    '''
    The method returns a string of the current model type.

    Args:
        model (RSTAB Class): Model Instance
    '''

    return model.clientModel.service.get_model_type()

def NewModelAsCopy(old_model_name: str = '',
                   old_model_folder: str = ''):
    '''
    The method creates a new model as copy from an existing model

    Args:
        old_model_name (str): Old Model Name
        old_model_folder (str): Old Model Folder
    '''

    # Old Model Name
    new_model_name = ''
    if '.rs9' in old_model_name:
        new_model_name = old_model_name[:-4] + '_copy'

    else:
         assert TypeError('Model ' + old_model_name +  ' does not exist')

    old_model_path = os.path.join(old_model_folder, old_model_name)

    # New Model Name
    newModelAsCopy = connectionGlobals.client.service.new_model_as_copy(new_model_name, old_model_path)

    return newModelAsCopy

def GetModelMainParameters(model = Model):
    '''
    The method returns the main parameters of the current model.

    Args:
        model (RSTAB Class, optional): Model to be edited
    '''

    # Client Model | Get Model Main Parameters
    return model.clientModel.service.get_model_main_parameters()

def GetModelId(model = Model):
    '''
    This method returns model id as a string.

    Args:
        model (RSTAB Class, optional): Model to be edited
    '''

    # Client Model | Get Model ID
    return model.clientModel.service.get_model_main_parameters().model_id

def GetModelParameters(model = Model):
    '''
    This method retuns the parameters of the current model.

    Args:
        model (RSTAB Class, optional): Model to be edited
    '''

    # Client Model | Get Model Parameters
    return model.clientModel.service.get_model_parameters()

def GetModelSessionId(model = Model):
    '''
    This method returns model session id as a string.

    Args:
        model (RSTAB Class, optional): Model to be edited
    '''

    # Client Model | Get Session Id
    return model.clientModel.service.get_session_id()

def GetName():
    '''
    This method returns app name as a string.
    '''

    # Client Application | Get Information
    return connectionGlobals.client.service.get_information().name

def GetVersion():
    '''
    This method returns version as a string.
    '''

    # Client Application | Get Information
    return connectionGlobals.client.service.get_information().version

def GetLanguage():
    '''
    This method returns language as a string.
    '''

    # Client Application | Get Information
    return connectionGlobals.client.service.get_information().language_name

def GetAppSessionId():
    '''
    This method returns session id as a string.
    '''

    # Client Application | Get Session ID
    return connectionGlobals.client.service.get_session_id()

def getPathToRunningRSTAB():
    '''
    Find the path to the directory where RSTAB is currently running.
    This is helpful when using server version, because it can't process relative paths.
    '''
    import psutil
    rfem6 = False
    rfem6Server = False
    path = ''

    for p in psutil.process_iter(['name', 'exe']):
        if p.info['name'] == 'RSTAB9.exe':
            idx = p.info['exe'].find('bin')
            path = p.info['exe'][:idx]
        elif p.info['name'] == 'RSTAB9Server.exe':
            idx = p.info['exe'].find('bin')
            path = p.info['exe'][:idx]
        elif p.info['name'] == 'RFEM6.exe':
            rfem6 = True
        elif p.info['name'] == 'RFEM6Server.exe':
            rfem6Server = True

    if rfem6 or rfem6Server:
        raise ValueError('Careful! You are running RSTAB Python Client on RSTAB.')
    if not path:
        raise ValueError('Is it possible that RSTAB is not runnnning?')

    return path

def GetListOfOpenedModels():

    connectToServer()
    models = connectionGlobals.client.service.get_model_list()

    return models




from RSTAB.initModel import Model

def GetPartsListAllByMaterial(model = Model):
    '''
    Returns Parts List All By Material
    '''
    try:
        return model.clientModel.service.get_parts_list_all_by_material()
    except:
        model.clientModel.service.generate_parts_lists()
        return model.clientModel.service.get_parts_list_all_by_material()

def GetPartsListMemberRepresentativesByMaterial(model = Model):
    '''
    Returns Parts List Member Representatives By Material
    '''
    try:
        return model.clientModel.service.get_parts_list_member_representatives_by_material()
    except:
        model.clientModel.service.generate_parts_lists()
        return model.clientModel.service.get_parts_list_member_representatives_by_material()

def GetPartsListMemberSetsByMaterial(model = Model):
    '''
    Returns Parts List Member Sets By Material
    '''
    try:
        return model.clientModel.service.get_parts_list_member_sets_by_material()
    except:
        model.clientModel.service.generate_parts_lists()
        return model.clientModel.service.get_parts_list_member_sets_by_material()

def GetPartsListMembersByMaterial(model = Model):
    '''
    Returns Parts List Members By Material
    '''
    try:
        return model.clientModel.service.get_parts_list_members_by_material()
    except:
        model.clientModel.service.generate_parts_lists()
        return model.clientModel.service.get_parts_list_members_by_material()



from RSTAB.initModel import Model
from RSTAB.enums import ModelCheckGetOptionType, ModelCheckProcessOptionType

class ModelCheck():

    @staticmethod
    def GetIdenticalNodes(tolerance, model = Model):
        """
        Args:
            tolerance (float): Tolerance
        Returns:
            Identical Nodes Object Group
        """

        operation = ModelCheckGetOptionType.IDENTICAL_NODES.name
        object_groups = model.clientModel.service.model_check__get_object_groups_operation(operation, tolerance)

        return object_groups

    @staticmethod
    def DeleteUnusedNodes(tolerance, object_groups, model = Model):
        """
        Args:
            tolerance (float): Tolerance
            object_groups (dict): Object Groups of Identical Nodes
        """

        process = ModelCheckProcessOptionType.DELETE_UNUSED_NODES.name
        model.clientModel.service.model_check__process_object_groups_operation(process, tolerance, object_groups)

    @staticmethod
    def UniteNodes(tolerance, object_groups, model = Model):
        """
        Args:
            tolerance (float): Tolerance
            object_groups (dict): Object Groups of Identical Nodes
        """

        process = ModelCheckProcessOptionType.UNITE_NODES_AND_DELETE_UNUSED_NODES.name
        model.clientModel.service.model_check__process_object_groups_operation(process, tolerance, object_groups)

    @staticmethod
    def GetNotConnectedLines(tolerance, model = Model):
        """
        Args:
            tolerance (float): Tolerance
        Returns:
            Not Connected Lines Line Groups
        """

        operation = ModelCheckGetOptionType.CROSSING_LINES.name
        line_groups = model.clientModel.service.model_check__get_object_groups_operation(operation, tolerance)

        return line_groups

    @staticmethod
    def CrossLines(tolerance, line_groups, model = Model):
        """
        Args:
            tolerance (float): Tolerance
            line_groups (dict): Line Groups of Not Connected Lines
        """

        process = ModelCheckProcessOptionType.CROSS_LINES.name
        model.clientModel.service.model_check__process_object_groups_operation(process, tolerance, line_groups)

    @staticmethod
    def GetNotConnectedMembers(tolerance, model = Model):
        """
        Args:
            tolerance (float): Tolerance
        Returns:
            Not Connected Members Member Groups
        """

        operation = ModelCheckGetOptionType.CROSSING_MEMBERS.name
        member_groups = model.clientModel.service.model_check__get_object_groups_operation(operation, tolerance)

        return member_groups

    @staticmethod
    def CrossMembers(tolerance, member_groups, model = Model):
        """
        Args:
            tolerance (float): Tolerance
            member_groups (dict): Member Groups of Not Connected Members
        """

        process = ModelCheckProcessOptionType.CROSS_MEMBERS.name
        model.clientModel.service.model_check__process_object_groups_operation(process, tolerance, member_groups)

    @staticmethod
    def GetOverlappingLines(model = Model):
        """
        Returns:
            Overlapping Line Groups
        """

        operation = ModelCheckGetOptionType.OVERLAPPING_LINES.name
        overlapping_lines = model.clientModel.service.model_check__get_object_groups_operation(operation)

        return overlapping_lines

    @staticmethod
    def GetOverlappingMembers(model = Model):
        """
        Returns:
            Overlapping Member Groups
        """

        operation = ModelCheckGetOptionType.OVERLAPPING_MEMBERS.name
        overlapping_members = model.clientModel.service.model_check__get_object_groups_operation(operation)

        return overlapping_members



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import LoadWizardType

class CombinationWizard():
    def __init__(self,
                no: int = 1,
                name: str = '',
                static_analysis_settings: int = 1,
                stability_analysis_setting: int = 1,
                consider_imperfection_case: bool = True,
                generate_same_CO_without_IC: bool = True,
                initial_state_cases: list = None,
                structure_modification: int = 1,
                user_defined_action_combinations: bool = False,
                favorable_permanent_actions: bool = False,
                reduce_number_of_generated_combinations: bool = False,
                comment: str = '',
                params: dict = None,
                model = Model):
        """
        Args:
            no (int): Combination Wizard Tag
            name (str, optional): User-Defined Name
            static_analysis_settings (int): Assign Static Analysis Settings
            stability_analysis_setting (int): Assign Stability Analysis Setting
            consider_imperfection_case (bool): Enable/disable Inferpaction Case
            generate_same_CO_without_IC (bool): Enable/disable Generate same CO without IC
            initial_state_cases (list): Assign Initial State Cases
                [[case_object (int), definition_type (initial_state_definition_type enum)],... ]
            structure_modification (int): Structure Modification to be considered
            user_defined_action_combinations (bool): Enable/disable User defined Combinations
            favorable_permanent_actions (bool): Enable/diable Favourable Permanent Actions
            reduce_number_of_generated_combinations (bool): Enable/disable Reduce Number of Generated Combinations
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """
        #  Client Model | Combination Wizard
        clientObject = model.clientModel.factory.create('ns0:combination_wizard')

        # Clear all attributes
        clearAttributes(clientObject)

        # Set Combination Wizard no.
        clientObject.no = no

        # Combination Wizard name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # What to generate
        clientObject.generate_combinations = LoadWizardType.GENERATE_LOAD_COMBINATIONS.name

        # Statical Analysis Model
        clientObject.static_analysis_settings = static_analysis_settings

        # Setting stability analysisa
        if stability_analysis_setting:
            clientObject.has_stability_analysis = True
            clientObject.stability_analysis_settings = stability_analysis_setting

        # Setting imperfection case
        clientObject.consider_imperfection_case = consider_imperfection_case
        clientObject.generate_same_CO_without_IC = generate_same_CO_without_IC

        # Setting initial state
        if initial_state_cases:
            clientObject.consider_initial_state = True
            clientObject.initial_state_items = model.clientModel.factory.create('ns0:array_of_combination_wizard_initial_state_items')

            for i,j in enumerate(initial_state_cases):
                cwisi = model.clientModel.factory.create('ns0:combination_wizard_initial_state_items_row')
                cwisi.no = i+1
                cwisi.row = model.clientModel.factory.create('ns0:combination_wizard_initial_state_items')
                clearAttributes(cwisi.row)
                cwisi.row.case_object = initial_state_cases[i][0] # int
                cwisi.row.definition_type = initial_state_cases[i][1].name # initial_state_definition_type enum

                clientObject.initial_state_items.combination_wizard_initial_state_items.append(cwisi)

        # Setting structure modificication
        # WS for new structure modification not yet available
        if structure_modification:
            clientObject.structure_modification_enabled = True
            clientObject.structure_modification = structure_modification

        # Setting user defined action combinations
        if user_defined_action_combinations:
            clientObject.user_defined_action_combinations = user_defined_action_combinations

        # Setting favorable combinations
        if favorable_permanent_actions:
            clientObject.favorable_permanent_actions = favorable_permanent_actions

        # Reduce number of generated combinations
        if reduce_number_of_generated_combinations:
            clientObject.reduce_number_of_generated_combinations = reduce_number_of_generated_combinations

        # Comment
        if comment:
            clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Clearing unused attributes
        deleteEmptyAttributes(clientObject)

        # Setting the combination wizard
        model.clientModel.service.set_combination_wizard(clientObject)

    # Setting result combination
    @staticmethod
    def SetResultCombination(no: int = 1,
                name: str = '',
                stability_analysis_setting: int = 1,
                consider_imperfection_case: bool = False,
                generate_same_CO_without_IC: bool = True,
                user_defined_action_combinations: bool = False,
                favorable_permanent_actions: bool = False,
                generate_subcombinations_of_type_superposition: bool = False,
                comment: str = '',
                params: dict = None,
                model = Model
                ):
        """
        Args:
            no (int): Result Combination Wizard Tag
            name (str, optional): User-Defined Name
            stability_analysis_settings (int): Assign Stability Analysis Settings
            consider_imperfection_case (bool): Enable/disable Imperfection Case
            generate_same_CO_without_IC (bool): Enable/disable Generate same CO without IC
            user_defined_action_combinations (bool): Enable/disable User Defined Action Combinations
            favorable_permanent_actions (bool): Enable/disable Favorable Permanent Actions
            generate_subcombinations_of_type_superposition (bool): Enable/disable Subcombinations of Type Superposition
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """
        # Client Model | Combination Wizard
        clientObject = model.clientModel.factory.create('ns0:combination_wizard')

        # Clear all attributes
        clearAttributes(clientObject)

        # Set Combination Wizard no.
        clientObject.no = no

        # Combination Wizard name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # What to generate
        clientObject.generate_combinations = LoadWizardType.GENERATE_RESULT_COMBINATIONS.name

        # Setting stability analysisa
        if stability_analysis_setting:
            clientObject.has_stability_analysis = True
            clientObject.stability_analysis_settings = stability_analysis_setting

        # Setting imperfection case
        clientObject.consider_imperfection_case = consider_imperfection_case

        # Setting additional setting for imperfection case
        clientObject.generate_same_CO_without_IC = generate_same_CO_without_IC

        # Setting user defined action combinations
        if user_defined_action_combinations:
            clientObject.user_defined_action_combinations = user_defined_action_combinations

        # Setting favorable combinations
        if favorable_permanent_actions:
            clientObject.favorable_permanent_actions = favorable_permanent_actions

        # Setting addition setting for sub - combinations
        clientObject.generate_subcombinations_of_type_superposition = generate_subcombinations_of_type_superposition

        # Comment
        if comment:
            clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Clearing unused attributes
        deleteEmptyAttributes(clientObject)

        # Setting the combination wizard
        model.clientModel.service.set_combination_wizard(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString, ConvertStrToListOfInt
from RSTAB.enums import SetType, ObjectTypes

class MemberSet():
    def __init__(self,
                 no: int = 1,
                 members_no: str = '1 4 5 8 9 12 13 16 17 20 21 24',
                 member_set_type = SetType.SET_TYPE_GROUP,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Args:
            no (int): Member Set Tag
            members_no (str): Numbers of Members Contained Within Member Set
            member_set_type (enum): Member Set Type Enumeration
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Member Set
        clientObject = model.clientModel.factory.create('ns0:member_set')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Set No.
        clientObject.no = no

        # Members number
        clientObject.members = ConvertToDlString(members_no)

        # Member Set Type
        clientObject.set_type = member_set_type.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Set to client model
        model.clientModel.service.set_member_set(clientObject)

    @staticmethod
    def ContinuousMembers(
                          no: int = 1,
                          members_no: str = '1 4 5 8 9 12 13 16 17 20 21 24',
                          comment: str = '',
                          params: dict = None,
                          model = Model):

        '''
        Args:
            no (int): Member Set Tag
            members_no (str): Numbers of Members Contained Within Continuous Member Set
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Member Set
        clientObject = model.clientModel.factory.create('ns0:member_set')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Set No.
        clientObject.no = no

        # Members number
        clientObject.members = ConvertToDlString(members_no)

        # Member Set Type
        clientObject.set_type = SetType.SET_TYPE_CONTINUOUS.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Set to client model
        model.clientModel.service.set_member_set(clientObject)

    @staticmethod
    def GroupOfmembers(
                       no: int = 1,
                       members_no: str = '1 4 5 8 9 12 13 16 17 20 21 24',
                       comment: str = '',
                       params: dict = None,
                       model = Model):

        '''
        Args:
            no (int): Member Set Tag
            members_no (str): Numbers of Members Contained Within Group of Members Member Set
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Member Set
        clientObject = model.clientModel.factory.create('ns0:member_set')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Set No.
        clientObject.no = no

        # Members number
        clientObject.members = ConvertToDlString(members_no)

        # Member Set Type
        clientObject.set_type = SetType.SET_TYPE_GROUP.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Set to client model
        model.clientModel.service.set_member_set(clientObject)

    @staticmethod
    def DeleteMemberSet(members_no: str = '1 2', model = Model):

        '''
        Args:
            members_no (str): Numbers of Members Sets to be deleted
            model (RSTAB Class, optional): Model to be edited
        '''

        # Delete from client model
        for member in ConvertStrToListOfInt(members_no):
            model.clientModel.service.delete_object(ObjectTypes.E_OBJECT_TYPE_MEMBER_SET.name, member)

    @staticmethod
    def GetMemberSet(object_index: int = 1, model = Model):

        '''
        Args:
            obejct_index (int): Member Set Index
            model (RSTAB Class, optional): Model to be edited
        '''

        # Get Member Set from client model
        return model.clientModel.service.get_member_set(object_index)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import NodalReleaseReleaseLocation

class NodalRelease():
    def __init__(self,
                 no: int = 1,
                 nodes: str = '',
                 nodal_release_type: int = 1,
                 release_location = NodalReleaseReleaseLocation.RELEASE_LOCATION_ORIGIN,
                 released_members: str = '1',
                 deactivate_release: bool = False,
                 name: str = '',
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
         Args:
            no (int): Nodal Release Tag
            nodes (str): Assigned Nodes
            nodal_release_type (int): Nodale Release Type Number
            release_location (enums): Nodal Release Release Location Enumeration
            released_members (str): Assigned Released Members
            deactivate_release (bool): Activate/Deactivate Nodal Release
            name (str, optional): User Defined Nodal Release Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Nodal Release
        clientObject = model.clientModel.factory.create('ns0:nodal_release')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Nodal Release No.
        clientObject.no = no

        # Assigned Nodes
        clientObject.nodes = ConvertToDlString(nodes)

        # Nodal Release Type
        clientObject.nodal_release_type = nodal_release_type

        # Released Members
        clientObject.released_members = ConvertToDlString(released_members)

        # Nodal Release Location
        clientObject.release_location = release_location.name

        # Deactivate Release
        clientObject.deactivated = deactivate_release

        # Nodal Release User defined name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Nodal Release to client model
        model.clientModel.service.set_nodal_release(clientObject)
